from flask_douwa.routes import rpcapi
from flask_douwa.rpc.type import *
#from hex_utils.dataformat import dataformat

class GeneratorRpc:
    package = "generator_id"
    name = "Generator"

    @rpcapi(returns=StringField)
    #@dataformat
    def GeneratorId(self):
        pass



import logging

import sys
import inspect

_py2 = sys.version_info[0] == 2
logger = logging.getLogger("rpc")


def rpcapi(returns, **parameter_types):
    """rpc注册装饰器"""

    def decorator(f):
        # Put the rule cache on the method itself instead of globally
        if not hasattr(f, '_rule_cache') or f._rule_cache is None:
            f._rule_cache = {f.__name__: [(returns, parameter_types)]}
        elif not f.__name__ in f._rule_cache:
            f._rule_cache[f.__name__] = [(returns, parameter_types)]
        else:
            f._rule_cache[f.__name__].append((returns, parameter_types))

        return f

    return decorator


def get_interesting_members(cls):
    """Returns a list of methods that can be routed to"""

    predicate = inspect.ismethod if _py2 else inspect.isfunction
    all_members = inspect.getmembers(cls, predicate=predicate)
    return [member for member in all_members
            if ((hasattr(member[1], "__self__") and not member[1].__self__ in inspect.getmro(cls)) if _py2 else True)
            and not member[0].startswith("_")]



def register(handlers):
    test2 = 'syntax = "proto3";\npackage {};\n'.format(handlers.package)
    service = "service {} {{\n".format(handlers.name)

    def make_proxy_method(name):

        i = handlers()
        view = getattr(i, name)
        return view

    members = get_interesting_members(handlers)

    for name, value in members:
        proxy = make_proxy_method(name)
        for idx, cached_rule in enumerate(value._rule_cache[name]):
            returns, method = cached_rule
            return_type = returns.name
            message = "message {}Request {{".format(name)
            c = 1
            paramenter_types = str()
            for i in method:
                paramenter_types = "{} {} = {};".format(method[i].name, i, c)
                message += paramenter_types
                c += 1
            message += "}\n"
            test2 += message
            message2 = "message {}Reply {{".format(name)
            message2 += "{} {} = 1;}}\n".format(return_type, "_result")
            test2 += message2
            service += "rpc {} ({}Request) returns ({}Reply){{}}\n".format(name, name, name)
    service += "}"
    test2 += service
    # logging.debug("proto file:{}".format(test2))
    return test2,handlers()







# -*- coding=utf8 -*-
"""protofuf 对象定义"""


class Field(object):
    """字段类型父类"""

    def __init__(self, name, index, number, label,
                 message_type=None,):
        self.name = name
        self.index = index
        self.number = number
        self.label = label
        self.message_type = message_type


class BoolField(Field):
    """bool类型"""
    type = 8
    cpp_type = 7
    default_value = False


class StringField(Field):
    """string类型"""
    type = 9
    cpp_type = 9
    default_value = u''


class BytesField(Field):
    """bytes类型"""
    type = 12
    cpp_type = 9
    default_value = ''


class DoubleField(Field):
    """double类型"""
    type = 1
    cpp_type = 5
    default_value = 0.0


class FloatField(Field):
    """float类型"""
    type = 2
    cpp_type = 6
    default_value = 0.0


class Int32Field(Field):
    """int32类型"""
    type = 5
    cpp_type = 1
    default_value = 0


class Int64Field(Field):
    """int64类型"""
    type = 3
    cpp_type = 2
    default_value = 0


class Uint32Field(Field):
    """uint32类型"""
    type = 13
    cpp_type = 3
    default_value = 0


class Uint64Field(Field):
    """uint64类型"""
    type = 4
    cpp_type = 4
    default_value = 0


class Sint32Field(Field):
    """sint32类型"""
    type = 17
    cpp_type = 1
    default_value = 0


class Sint64Field(Field):
    """sint64类型"""
    type = 18
    cpp_type = 2
    default_value = 0


class Fixed32Field(Field):
    """fixed32类型"""
    type = 7
    cpp_type = 3
    default_value = 0


class Fixed64Field(Field):
    """fixed32类型"""
    type = 6
    cpp_type = 4
    default_value = 0


class Sfixed32Field(Field):
    """sfixed32类型"""
    type = 15
    cpp_type = 1
    default_value = 0


class Sfixed64Field(Field):
    """sfixed64类型"""
    type = 16
    cpp_type = 2
    default_value = 0


class MessageField(Field):
    """message类型"""
    type = 11
    cpp_type = 10
    default_value = None


class MapField(Field):
    """map类型"""
    type = 11
    cpp_type = 10
    default_value = []

    def __init__(self, name, index, number, label,
                 message_type=None, key_type=None, value_type=None):
        self.key_type = key_type
        self.value_type = value_type
        super(MapField, self).__init__(
            name, index, number, label, message_type)


field_map = {
    'bool': BoolField,
    'string': StringField,
    'bytes': BytesField,
    'double': DoubleField,
    'float': FloatField,
    'int32': Int32Field,
    'int64': Int64Field,
    'uint32': Uint32Field,
    'uint64': Uint64Field,
    'sint32': Sint32Field,
    'sint64': Sint64Field,
    'fixed32': Fixed32Field,
    'fixed64': Fixed64Field,
    'sfixed32': Sfixed32Field,
    'sfixed64': Sfixed64Field,
}


class ServiceMethod(object):
    """service rpc method"""
    def __init__(self, name, request_type, response_type):
        self.name = name
        self.request_type = request_type
        self.response_type = response_type


class Service(object):
    """service"""
    def __init__(self, name, methods=None):
        self.name = name
        self.methods = methods or {}


class Message(object):
    """message"""
    def __init__(self, name, fields=None):
        self.name = name
        self.fields = fields or {}


class Protobuf(object):
    """解析的对象"""

    def __init__(self, package='', syntax='proto3'):
        self.package = package
        self.syntax = syntax

        self.services = {}
        self.messages = {}
        self.filename = None



# # -*- coding: utf-8 -*-
#
# from django.utils.translation import ugettext_lazy as _
#
# # Sequence Field
#
# SEQUENCE_KEY = _(u'Key')
# SEQUENCE_VALUE = _(u'Value')
# SEQUENCE_TEMPLATE = _(u'Template')
# SEQUENCE_CREATED = _(u'Created at')
# SEQUENCE_UPDATED = _(u'Updated at')
#
# SEQUENCE_MODEL_NAME = _(u'Sequence')
# SEQUENCE_MODEL_NAME_PLURAL = _(u'Sequences')
#
# SEQUENCE_FIELD_DESCRIPTION = _(u'Templated sequence object')
#
# # Errors
#
# SEQUENCE_FIELD_PATTERN_MISMATCH = _(
#     u'The value does not match the specified pattern'
# )
#
# SEQUENCE_FIELD_MISSING_KEY = _(
#     u'The key parameter is mandatory and is missing'
# )
#
# SEQUENCE_FIELD_KEY_MISMATCH = _(
#     u'The key \'%(key)s\' does not match any of the existing '
#     u'sequence model objects'
# )



# Sequence Field constants

SEQUENCE_KEY_LENGTH = 128

SEQUENCE_TEMPLATE_LENGTH = 128

SEQUENCE_DEFAULT_VALUE = 1

SEQUENCE_FIELD_ADMIN = True

SEQUENCE_FIELD_DEFAULT_PATTERN = r'(\d+)'

SEQUENCE_FIELD_DEFAULT_TEMPLATE = '%N'

SEQUENCE_FIELD_DEFAULT_EXPANDERS = (
    'sequence_field.expanders.NumericExpander',
    'sequence_field.expanders.TimeExpander',
    'sequence_field.expanders.ParameterExpander',
)



# -*- coding=utf8 -*-
import logging
import os
import types
from ply import (
    yacc,
    lex,
)
from google.protobuf import (
    descriptor as _descriptor,
    message as _message,
    reflection as _reflection,
    symbol_database as _symbol_database
)
import grpc

from .lexer import *
from .grammar import *

logger = logging.getLogger("rpc")


def load_byte(data, lexer=None, parser=None):
    """解析protobuf文件

    Args:
        proto_path (string): protobuf文件路径
        lexer (Lexer): 词法分析
        parser (Parser): 语法分析器
    """
    if not lexer:
        lexer = lex.lex()
    if not parser:
        parser = yacc.yacc(debug=False, write_tables=0)

    lexer.lineno = 1
    result = parser.parse(data)
    result.filename = "haha.proto"
    return result

def load(proto_path, lexer=None, parser=None):
    """解析protobuf文件

    Args:
        proto_path (string): protobuf文件路径
        lexer (Lexer): 词法分析
        parser (Parser): 语法分析器
    """
    proto_path = os.path.abspath(proto_path)

    if not proto_path.endswith('.proto'):
        raise Exception('file name must end with .proto')

    with open(proto_path, 'r') as pf:
        data = pf.read()

    if not lexer:
        lexer = lex.lex()
    if not parser:
        parser = yacc.yacc(debug=False, write_tables=0)

    lexer.lineno = 1
    result = parser.parse(data)
    result.filename = proto_path
    return result


def transform(proto):
    """把load生成的对象转换为原生对象

    Args:
        proto (Protobuf): load返回对象
    """
    file_name = os.path.basename(proto.filename)
    module_name = file_name.replace('.proto', '_pb2')
    proto_module = types.ModuleType(module_name)

    proto_module._sym_db = _symbol_database.Default()
    descriptor = _descriptor.FileDescriptor(
        name=file_name,
        package=proto.package,
        syntax=proto.package,
        serialized_pb=''
    )
    proto_module._sym_db.RegisterFileDescriptor(descriptor)

    # serialized_start和serialized_end真不知道怎么算出来的, 这里就简单的估算一下
    serialized_start = serialized_end = 23 + len(proto.package)
    descriptor_map = {}
    for message_name, message in proto.messages.items():
        fields = []
        for field_name, field in message.fields.items():
            if fields:
                serialized_end += 15
            else:
                serialized_end += 25

            fields.append(_descriptor.FieldDescriptor(
                name=field_name,
                full_name='.'.join((proto.package, message_name, field_name)),
                index=field.index,
                number=field.number,
                type=field.type,
                cpp_type=field.cpp_type,
                label=field.label,
                has_default_value=False,
                default_value=field.default_value,
                message_type=None,
                enum_type=None,
                containing_type=None,
                is_extension=False,
                extension_scope=None,
                options=None,
            ))

        descriptor_map[message_name] = _descriptor.Descriptor(
            name=message_name,
            full_name='.'.join((proto.package, message_name)),
            filename=None,
            file=descriptor,
            containing_type=None,
            fields=fields,
            extensions=[],
            nested_types=[],
            enum_types=[],
            options=None,
            is_extendable=False,
            syntax=proto.syntax,
            extension_ranges=[],
            oneofs=[],
            serialized_start=serialized_start,
            serialized_end=serialized_end,
        )
        serialized_start = serialized_end = serialized_end + 2

    for descriptor_name, des in descriptor_map.items():
        descriptor.message_types_by_name[descriptor_name] = des
        setattr(proto_module, descriptor_name,
                _reflection.GeneratedProtocolMessageType(
                    descriptor_name,
                    (_message.Message,),
                    {
                        'DESCRIPTOR': des,
                        '__module__': module_name
                    }
                ))
        proto_module._sym_db.RegisterMessage(
            getattr(proto_module, descriptor_name))

    return proto_module


def make_client(proto):
    """根据protobuf文件直接生成客户端相关对象"""
    #proto = load(proto_path)
    proto_module = transform(proto)

    for service_name, service in proto.services.items():
        class_name = '{}Stub'.format(service_name)

        def init_method(self, channel):
            for method_name, method in service.methods.items():
                setattr(self, method_name, channel.unary_unary(
                    '/{}/{}'.format('.'.join((proto.package, service_name)),
                                    method_name),
                    request_serializer=getattr(
                        proto_module, method.request_type).SerializeToString,
                    response_deserializer=getattr(
                        proto_module, method.response_type).FromString,
                ))

        stub_class = type(class_name, (object,), {
            '__module__': proto_module.__name__,
            '__init__': init_method
        })
        setattr(proto_module, class_name, stub_class)

    return proto_module

def bind(proto, servicer):
    """根据protobuf文件直接生成服务端相关对象"""

    proto_module = transform(proto)
    rpc_method_handlers = dict()

    for service_name, service in proto.services.items():

        def init_method():
            for method_name, method in service.methods.items():
                from functools import partial
                import inspect

                def simple(fun, return_fun, request, context):
                    args = inspect.getargspec(fun)[0]
                    kwargs = dict()
                    for i in args[1:]:
                        kwargs[i] = getattr(request,i)
                    _result = fun(**kwargs)
                    logger.debug("args:{}".format(kwargs))
                    logger.debug("result:{}".format(_result))
                    return return_fun(_result=_result)
                rpc_method_handlers[method_name] = grpc.unary_unary_rpc_method_handler(
                partial(simple,getattr(servicer, method_name),getattr(proto_module, method.response_type)),
              request_deserializer=getattr(
                        proto_module, method.request_type).FromString,
              response_serializer=getattr(
                        proto_module, method.response_type).SerializeToString,
                )

        init_method()
            # '.'.join((proto.package + str(hex_api.config.HEX_PARTNER_CODE), service_name)),
    generic_handler = grpc.method_handlers_generic_handler(
            '.'.join((proto.package , service_name)),
            rpc_method_handlers)
    return generic_handler




# -*- coding: utf-8 -*-

# Sequence Field Expanders

import re

class BaseExpander(object):

    def __init__(self, template=None, count=None, params={}, value=None):
        self.template = template
        self.count = count
        self.params = params
        self.value = value

    def setvars(self, template=None, count=None, params={}, value=None):
        # Parameters take precedence over attributes
        template = template if template is not None else self.template
        count = count if count is not None else self.count
        params = params if len(params)>0 else self.params
        value = value if value is not None else self.value
        # Copy current value from template
        if value is None or len(value) == 0:
            value = template
        return (template, count, params, value, )

    def expand(self, template=None, count=None, params={}, value=None):
        # Do nothing, just return the value
        (template, count, params, value) = self.setvars(
            template, count, params, value
        )
        return value


class NumericExpander(BaseExpander):

    regexp = r'^(.*)(?P<numeric_placeholder>%N+)(.*)$'

    def __init__(self, template=None, count=None, params={}, value=None):
        super(NumericExpander, self).__init__(template, count, params, value)
        self.format = '%d'

    def set_format_from_numeric_placeholder(self, numeric_placeholder):
        format = '%%0%dd' % len(numeric_placeholder.replace('%', ''))
        self.format = format

    def expand_count(self, count=None):
        if count is None:
            count = self.count
        return self.format % count

    def get_numeric_placeholder_from_groupdict(self, groupdict):
        if 'numeric_placeholder' in groupdict:
            return groupdict['numeric_placeholder']
        else:
            return ''

    def get_numeric_placeholder(self, template=None):
        if template is None:
            template = self.template
        match = re.match(self.regexp, template)
        if match is not None:
            return self.get_numeric_placeholder_from_groupdict(
                match.groupdict()
            )
        else:
            return None

    def expand_with_expanded_count(self, expanded_count, value=None):
        if value is None:
            value = self.value
        parts = [
            re.sub(self.regexp, r'\1', value),
            expanded_count,
            re.sub(self.regexp, r'\3', value)
        ]
        return ''.join(parts)

    def expand(self, template=None, count=None, params={}, value=None):
        (template, count, params, value) = self.setvars(
            template, count, params, value
        )
        numeric_placeholder = self.get_numeric_placeholder()
        self.set_format_from_numeric_placeholder(numeric_placeholder)
        expanded_count = self.expand_count()
        value = self.expand_with_expanded_count(expanded_count, value)
        return value


class ParameterExpander(BaseExpander):

    def expand(self, template=None, count=None, params={}, value=None):
        (template, count, params, value) = self.setvars(
            template, count, params, value
        )
        return value.format(**params)


class TimeExpander(BaseExpander):

    def expand(self, template=None, count=None, params={}, value=None):
        import time
        (template, count, params, value) = self.setvars(
            template, count, params, value
        )
        return time.strftime(value)



from flask_sqlalchemy import SQLAlchemy
from .constants import *
from .utils import *
import datetime
from app import db


def get_session():
    return db.create_scoped_session(options={'autocommit': True, 'autoflush': False})


class Sequence(db.Model):
    id = db.Column(db.Integer, autoincrement=True, primary_key=True)
    key = db.Column(db.String(50), unique=True)
    value = db.Column(db.Integer, nullable=True)
    template = db.Column(db.String(50), nullable=True)
    created_time = db.Column(db.DateTime, nullable=True)
    update_time = db.Column(db.DateTime, nullable=True)

    def __repr__(self):
        return '<Sequence {}>'.format(self.key)

    def increment(self, commit=True):
        # sequence_session = get_session()
        self.value += 1
        self.update_time = datetime.datetime.now()
        if commit:
            db.session.flush()

    def next_value(self, template=None, params=None,
                   expanders=None, commit=True):

        default_template = self.template

        default_expanders = SEQUENCE_FIELD_DEFAULT_EXPANDERS

        count = self.value
        template = template if template is not None else default_template
        params = params if params is not None else {}
        expanders = expanders if expanders is not None else default_expanders
        if commit:
            self.increment()
        return expand(template, count, params, expanders=expanders)

    @classmethod
    def create_if_missing(cls, key, template=None):
        default_template = SEQUENCE_FIELD_DEFAULT_TEMPLATE
        try:
            # sequence_session=get_session()
            seq = db.session.query(cls).filter_by(key=key).first()
            if seq:
                return seq
            else:
                seq = Sequence()
                seq.value=SEQUENCE_DEFAULT_VALUE
                seq.key = key
                seq.template = template or default_template
                seq.created_time = datetime.datetime.now()
                db.session.add(seq)
                db.session.commit()
                return seq
        except Exception:
            return None

    @classmethod
    def next(cls, key, template=None, params=None,
             expanders=None, commit=True):
        seq = Sequence.create_if_missing(key, template)
        return seq.next_value(template, params, expanders, commit)

    @classmethod
    def get_template_by_key(cls, key):
        default_template = SEQUENCE_FIELD_DEFAULT_TEMPLATE
        try:
            db_session=get_session()
            seq = db_session.query(Sequence).filter_by(key=key).first()
            return seq.template
        except Exception:
            return default_template



from setuptools import setup


setup(
    name='Flask-Douwa',
    version='0.4.4',
    license='MIT',
    author='jhchen',
    author_email='watasihakamidesu@sina.cn',
    description=u'一个flask插件',
    long_description=__doc__,
    packages=['flask_douwa', 'flask_douwa.rpc', 'flask_douwa.protoparser', 'sequence_field'],
    zip_safe=False,
    include_package_data=True,
    platforms='any',
    install_requires=[
        'Flask',
        'grpcio==1.7.0',
        'protobuf==3.4.0',
        'ply==3.10',
        'msgpack-python==0.4.8',
        'confluent-kafka'

        #'kafka-python==1.3.6.dev0',
    ],
    classifiers=[
        'Framework :: Flask',
        'Natural Language :: English',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ]
)



# -*- coding: utf-8 -*-

from sqlalchemy import NVARCHAR, TypeDecorator
from .models import Sequence
from .constants import SEQUENCE_FIELD_DEFAULT_EXPANDERS


# 这里面的初始化的时候的部分属性，我研究了好久，也每看出是干啥的来，
class SequenceField(TypeDecorator):
    """ Stores sequence values based on templates. """
    impl = NVARCHAR

    def __init__(self, *args, **kwargs):
        if kwargs.get('key'):
            self.key = kwargs.pop('key')
            # 默认的模板
            default_template = Sequence.get_template_by_key(self.key)
            self.template = kwargs.pop('template', default_template)
            Sequence.create_if_missing(self.key, self.template)
            # 默认的数据模板解释器
            default_expanders = SEQUENCE_FIELD_DEFAULT_EXPANDERS
            # 模板中需要替换的参数
            self.params = kwargs.pop('params', {})
            self.expanders = kwargs.pop('expanders', default_expanders)
            # 控制是否自动生成，如果为FASLE这个sequence就不会自动生成
            self.auto = kwargs.pop('auto', False)
        super(SequenceField, self).__init__(*args, **kwargs)

    def _next_value(self):
        # 根据key查询是否是已经创建过的序列，主要目的是取得关键性的数据
        seq = Sequence.create_if_missing(self.key, self.template)
        return seq.next_value(self.template, self.params, self.expanders)

    def process_bind_param(self, value, dialect):
        if self.auto and not value:
            sequence_string = self._next_value()
            return sequence_string
        elif self.auto and isinstance(value, dict):
            self.params = value
            self.key = str(value)
            sequence_string = self._next_value()
            return sequence_string
        else:
            return value

    def process_result_value(self, value, dialect):
        return value
